# sb::Driver::Text
# == Author(s) : Takuya Otani <takuya.otani@gmail.com> ==
# == Copyright (C) 2005 SimpleBoxes/SerendipityNZ Ltd. ==

# ======================================
package sb::Driver::Text;
use strict;
use sb::Lock ();
use sb::Language ();
use sb::Driver ();
use base qw( sb::Driver );
use vars qw( $VERSION );
$VERSION = '0.21';
# 0.21 [2009/01/31] modified _load_list to make it fast
# 0.20 [2008/11/09] added get_nextid
# 0.19 [2008/11/07] renamed load_ids, save_ids / added create_id_table
# 0.18 [2008/11/04] modified setup/_encode to make them consistent with other classes
# 0.17 [2008/07/20] added alter
# 0.16 [2007/12/04] added insert / modified setup to create a lock file
# 0.15 [2007/08/07] changed load and _load_list to add "limit" in %param
# 0.14 [2007/08/04] applied some optimazations
# 0.13 [2007/07/31] changed create_table to remove an unused variable
# 0.12 [2007/07/16] changed return to make them consistent
# 0.11 [2007/07/11] changed _save_list and _load_list to handle huge data
# 0.10 [2007/01/06] changed _sort_object/_encode to make them faster
# 0.09 [2006/03/31] added driver_param
# 0.08 [2006/03/29] added count
# 0.07 [2006/03/17] changed bracket rule
# 0.06 [2006/03/02] replaced _name into Driver, added create_table
# 0.05 [2006/02/26] removed matched_number to move base class
# 0.04 [2005/11/24] changed _save_ids to save index file correctly
# 0.03 [2005/10/28] refactored
# 0.02 [2005/08/23] changed _load_detail to separate loading an object
# 0.01 [2005/03/01] changed how to lock a file
# 0.00 [2005/02/02] generated
# --------------------------------------
# // defines / static variables
sub DEFAULT_DIRECTORY  (){ './data/' };
sub DEFAULT_SUFFIX     (){ '.cgi' };
sub DEFAULT_INDEX      (){ 'id' };
sub DEFAULT_PERMISSION (){ 0666 };
sub DEFAULT_LOCKFILE   (){ 'save' };
sub DEFAULT_UNIT       (){ 100 };
sub TEMPORARY_FILE     (){ '_tmp' };
# --------------------------------------
# // override base class - param
sub driver_param
{
	my $class = shift;
	return (
		'name'       => DEFAULT_DIRECTORY,
		'suffix'     => DEFAULT_SUFFIX,
		'index'      => DEFAULT_INDEX,
		'permission' => DEFAULT_PERMISSION,
		'lock_file'  => DEFAULT_LOCKFILE,
		'tmp_file'   => TEMPORARY_FILE,
		'unit'       => DEFAULT_UNIT,
	);
}
# --------------------------------------
# // override base class
sub setup
{
	my $self = shift;
	$self->SUPER::setup(
		$self->driver_param(),
		'max' => {}, # buffer for id index
		@_,
	);
	my $lock = sb::Lock->basedir . $self->{'lock_file'};
	if (!-e $lock)
	{
		require sb::File;
		sb::File->new('fullpath'=>$lock)->touch($self->{'permission'});
	}
	$self->load_ids();
	return $self;
}
# --------------------------------------
# // override base class - public method
sub load($$;@)
{
	my $self  = shift;
	my $class = shift; # sets class name of sb::Data::Object
	my %param = ( # input paramaeters
		'cond'  => {},    # searching option
		'id'    => undef, # inidicates id to load directly
		'logic' => 'and', # logic factor for searching option
		'bgn'   => 0,     # offset of objects to load
		'num'   => -1,    # maximum of objects to load
		'sort'  => 'id',  # sorting option
		'order' => 0,     # sorting order 0:descent 1:ascent
		'limit' => undef, # parameter limitation
		@_
	);
	my $func = _generate_condition($param{'logic'},%{$param{'cond'}});
	my @array = $self->_load_list($class,$func,%param);
	$self->{'num'} = $#array + 1;
	if ( $self->{'num'} > 1 )
	{ # there are more than one objects
		@array = _sort_object($param{'sort'},$param{'order'},@array);
		@array = splice(@array,$param{'bgn'},$param{'num'}) if ( $param{'num'} > 0 and @array > $param{'num'} );
	}
	return @array;
}
sub count($$;@)
{
	my $self = shift;
	# sb::Driver::Text does not have proper count method.
	# So we should call "load" to count matched objects.
	$self->load(@_);
	return $self->matched_number();
}
sub save($@)
{
	my $self = shift;
	my @objs = @_; # set instances of sb::Data::Object to save
	return if ( $#objs == -1 );
	my $class = ref($objs[0]);
	my $lock = sb::Lock->basedir . $self->{'lock_file'};
	@objs = sort { $a->id <=> $b->id } @objs;
	my $lfh = sb::Lock->locked_open($lock) or die(sb::Language->get->string('Failed to lock.'));
	$self->_save_list($class,@objs);
	$self->save_ids();
	close($lfh);
	return 1;
}
sub insert($$)
{
	my $self = shift;
	my $obj = shift; # set an instance of sb::Data::Object to save
	my $class = ref($obj);
	if ( !defined($self->{'max'}->{$self->name($class)})
	  or $obj->id >= $self->{'max'}->{$self->name($class)} )
	{
		$self->{'max'}->{$self->name($class)} = $obj->id + 1;
	}
	$self->save($obj);
	return 1;
}
sub new_id($$)
{
	my $self = shift;
	my $class = shift; # set class name of sb::Data::Object
	return $self->{'max'}->{$self->name($class)}++;
}
sub create_table($$)
{
	my $self  = shift;
	my $class = shift; # set class name of sb::Data::Object
	my $file = $self->{'name'} . $self->name($class) . '0' . $self->{'suffix'};
	require sb::File;
	sb::File->new('fullpath'=>$file)->touch($self->{'permission'});
	return 1;
}
sub alter($$$;$)
{
	my $self = shift;
	my $class = shift; # sets class name of sb::Data::Object
	my $target = shift;
	my $mode = shift; # CHANGE / ADD / DROP
	$mode = 'CHANGE' if ( $mode ne 'ADD' and $mode ne 'DROP' );
	return 1;
}
sub get_nextid($$)
{
	my $self = shift;
	my $class = shift; # sets class name of sb::Data::Object
	return 0 if ( !defined( $self->{'max'}->{$self->name($class)} ) );
	return $self->{'max'}->{$self->name($class)};
}
# --------------------------------------
# // public functions - id table
sub load_ids
{
	my $self = shift;
	my $file = $self->{'name'} . $self->{'index'} . $self->{'suffix'};
	if ( !-e $file )
	{
		return $self->create_id_table();
	}
	open(IDFILE_IN,"<$file") or die(sb::Language->get->string('Failed to open the file') . ' : ' . $file);
	while (my $line = <IDFILE_IN>)
	{
		$line =~ tr/\x0D\x0A//d;
		my ($key,$val) = split("\t",$line,2);
		next if ($key eq '');
		$self->{'max'}->{$key} = int($val);
	}
	close(IDFILE_IN);
	return 1;
}
sub save_ids
{
	my $self = shift;
	my $file = $self->{'name'} . $self->{'index'} . $self->{'suffix'};
	open(IDFILE_OUT,">$file") or die(sb::Language->get->string('Failed to open the file') . ' : ' . $file);
	binmode(IDFILE_OUT);
	foreach my $key ( keys(%{$self->{'max'}}) )
	{
		print IDFILE_OUT $key,"\t",$self->{'max'}->{$key},"\n";
	}
	close(IDFILE_OUT);
	return 1;
}
sub create_id_table
{
	my $self = shift;
	my $file = $self->{'name'} . $self->{'index'} . $self->{'suffix'};
	if ( !-e $file )
	{
		require sb::File;
		return sb::File->new('fullpath'=>$file)->touch($self->{'permission'});
	}
	return 1;
}
# --------------------------------------
# // private functions - instance method
sub _load_list
{
	my $self  = shift;
	my $class = shift; # sets class name of sb::Data::Object
	my $func  = shift; # sets filter function
	my %param = @_;
	my @array = ();
	my $is_id = (defined($param{'id'}));
	my $name  = $self->name($class);
	my $bgn = ($is_id) 
		? int($param{'id'} / $self->{'unit'}) 
		: 0;
	my $end = ($is_id) 
		? int($param{'id'} / $self->{'unit'}) 
		: int($self->{'max'}->{$name} / $self->{'unit'});
	my @elements = $self->param_list($class,$param{'limit'});
	for my $i ($bgn .. $end)
	{
		my $file = $self->{'name'} . $name . $i . $self->{'suffix'};
		next if (!-r $file or -z $file);
		open(DATAIN,"<$file") or die(sb::Language->get->string('Failed to open the file') . ' : ' . $file);
		while (my $line = <DATAIN>)
		{ # load each object to allocate
			$line =~ tr/\x0D\x0A//d;
			my @data = _decode($line);
			next if ($is_id and $param{'id'} != $data[0]);
			my $obj = $class->new();
			map { $obj->{$_} = shift(@data) } @elements;
			next if ( &{$func}($obj) );
			push(@array,$obj);
			last if ($is_id);
		}
		close(DATAIN);
	}
	return @array;
}
sub _save_list
{
	my $self  = shift;
	my $class = shift; # set class name of sb::Data::Object
	my $name  = $self->name($class);
	my %objs  = ();
	foreach my $obj (@_)
	{
		next if ($obj->id eq '');
		next if (ref($obj) ne $class);
		my $id = int($obj->id / $self->{'unit'});
		if ( !$objs{$id} )
		{
			$objs{$id} = [$obj];
		}
		else
		{
			push(@{$objs{$id}},$obj);
		}
	}
	my $temp = $self->{'name'} . $name . $self->{'tmp_file'};
	foreach my $id ( keys %objs )
	{
		open(LISTOUT,">$temp") or die(sb::Language->get->string('Failed to open the file') . ' : ' . $temp);
		binmode(LISTOUT);
		my @target = @{$objs{$id}};
		my $file = $self->{'name'} . $name . $id . $self->{'suffix'};
		if (-e $file)
		{
			open(LISTIN,"<$file") or die(sb::Language->get->string('Failed to open the file') . ' : ' . $file);
			while (my $line = <LISTIN>)
			{ # load each object to check updating
				my @data = _decode($line);
				if (@target and $data[0] == $target[0]->id)
				{
					print LISTOUT _encode($target[0],$name) if (!$target[0]->erased);
					shift(@target);
				}
				else
				{
					print LISTOUT $line;
				}
			}
			close(LISTIN);
		}
		foreach my $obj ( @target )
		{
			next if ($obj->erased);
			print LISTOUT _encode($obj,$name);
		}
		close(LISTOUT);
		rename($temp,$file);
		chmod($self->{'permission'},$file);
	}
}
# --------------------------------------
# // private functions - tools
sub _generate_condition
{
	my $logic = shift;
	my %cond  = @_;
	return (sub { return(0) }) if (! keys(%cond) );
	my %reg = ();
	foreach my $key ( keys(%cond) )
	{
		next if ( $key eq '__range' or $key eq '__combo' );
		if ( $cond{$key} =~ /^\/(.*)\/$/ )
		{
			$reg{$key} = $1;
			$cond{$key} = undef;
		}
	}
	my $re = ( keys(%reg) ) ? 1 : 0;
	my $func  = sub {
		my $flag = 0;
		foreach my $key ( keys(%reg) )
		{
			my $word = $reg{$key};
			$flag = ( $_[0]->{$key} =~ /\Q$word\E/i );
			last if ($flag);
		}
		return (!$flag) if ( ($flag and $logic eq 'or' and $re) or (!$flag and $logic eq 'and' and $re) );
		foreach my $key ( keys(%cond) )
		{
			next if ( $key eq '__range' or $key eq '__combo' );
			next if (! defined($cond{$key}) or ! defined($_[0]->{$key}));
			if ( ref($cond{$key}) eq 'ARRAY' )
			{
				my $check = $_[0]->{$key};
				$check += sb::Time->diff_timezone($_[0]->{'tz'}, 0) if ( $cond{'__range'}->{$key} eq 'tz' );
				$flag = (defined($cond{'__range'}->{$key})) 
					? ( $check >= $cond{$key}[0] and $check < $cond{$key}[1] )
					: ( grep( /^\Q$check\E$/, @{$cond{$key}} ) );
			}
			else
			{
				$flag = ( $_[0]->{$key} eq $cond{$key} );
			}
			if (defined($cond{'__combo'}->{$key}) and !$flag)
			{
				my %combo = %{$cond{'__combo'}};
				$flag = ( index($_[0]->{$combo{$key}},$combo{$combo{$key}}) > -1 );
			}
			last if ( ($flag and $logic eq 'or') or (!$flag and $logic eq 'and') );
		}
		return (!$flag);
	};
	return $func;
}
sub _sort_object
{
	my $elem  = shift;
	my $order = shift;
	my @array = @_;
	my (@data,@ids);
	foreach (@array)
	{
		push(@data, $_->{$elem});
		push(@ids,  $_->{'id'});
	}
	return @array[ sort {
		$data[$b] <=> $data[$a] or $data[$a] cmp $data[$b] or $ids[$b] <=> $ids[$a]
	} 0 .. $#ids ] if ($order);
	return @array[ sort {
		$data[$a] <=> $data[$b] or $data[$b] cmp $data[$a] or $ids[$a] <=> $ids[$b]
	} 0 .. $#ids ];
}
sub _decode
{
	return map { s/\\(.)/$1 eq 't' and "\t" or $1 eq 'n' and "\n" or "$1"/eg; $_; } split("\t",$_[0]);
}
sub _encode
{
	my ($obj,$name) = @_;
	my @data = ();
	foreach my $elem ( $obj->elements )
	{
		push(@data,&_linefeed($obj->{$elem}));
	}
	return join("\t",map { s/\\/\\\\/g; s/\t/\\t/g; s/\n/\\n/g; $_; } @data) . "\t\n";
}
sub _linefeed
{ # unify linefeed code
	$_[0] =~ s/\x0D\x0A/\n/g;
	$_[0] =~ tr/\x0D\x0A/\n\n/;
	while ($_[0] =~ /\n$/) { chomp($_[0]); }
	return $_[0];
}
1;
__END__
